import math
import random
import time
import os
from operator import itemgetter
import numpy as np
import networkx as nx

class NoInstance(Exception):
	pass

class EmptyMu(Exception):
	pass

class EmptyNu(Exception):
	pass

class FormatErrorMuNu(Exception):
	pass

class FormatError(Exception):
	pass

class Resonance(object):
	def __init__(self, value, sigma, name, adddeltalambda, fromfile=False):
		self.val=float(value)
		self.sigma=float(sigma)/2+adddeltalambda
		if fromfile:
			self.sigma=float(sigma)+adddeltalambda
		self.name=name
		self.lb=self.val-self.sigma
		self.ub=self.val+self.sigma
	def __str__(self):
		return "val:"+str(self.val)+"s:"+str(self.sigma)
	def interferes(self,resB):
		return not (self.lb>resB.ub or self.ub<resB.lb)
	def interferes(self,value,sigma):
		return not (self.lb>value+sigma or self.ub<value-sigma)

class ResonanceList(object):
	def __init__(self):
		self.reslist=[]
	def __str__(self):
		return ":".join([str(r)+":"+str(s) for r,s in self.reslist])
	def append(self, one):
		self.reslist.append(one)
	def getRes(self):
		return [r.val for r in self.reslist]
	def getResPert(self, LBlam, UBlam):
		out=[]
		a=random.uniform(0.0,1)
		for r in self.reslist:
			if r.lb+abs(r.lb-r.ub)*a > LBlam and r.lb+abs(r.lb-r.ub)*a < UBlam:
				out.append(r.lb+abs(r.lb-r.ub)*a)
		return out
	def getResInst(self):
		return [r for r in self.reslist]
	def getres(self, index):
		return self.reslist[index]
	def getval(self, index):
		return self.reslist[index].val
	def getsigma(self, index):
		return self.reslist[index].sigma
	def len(self):
		return len(self.reslist)
	def getSigmas(self):
		return [r.sigma for r in self.reslist]
	def get_interferences(self, res):
		out=[]
		for r in self.reslist:
			if r.interferes(res):
				out.append(r)
		return out
	def get_interferences(self, value, sigma):
		out=[]
		for r in self.reslist:
			if r.interferes(value, sigma):
				out.append(r.name)
		return out

class instance(object):
	def __init__(self, options):
		self._rows=[]
		
		self.minr=9999999999
		self.maxr=0
		self.nelements=0
		self.sizesmallest=0
		self.sizebiggest=0
		self.LBlam=options.lambdamin
		self.UBlam=options.lambdamax
		self.aspinstance=""

		self.name="Ropt_"+str(options.rmin)+"_"+str(options.rstep)+"_"+str(options.rmax)

		idc=0
		sigmaall=[]
		print("CREATING RADIUSES")
		if options.filename!=None:
			f=open(options.filename)
			lll=f.readlines()
			f.close()
			for ll in lll:
				ll=ll.split("\t")
				idc=ll[0]
				r=ll[1]
				ressigma=[(l.split(":")) for l in ll[5:]]
				self.addrowmatlab(options,ressigma,idc,r,options.rtol,True)
		else:
			try:
				import lambdaengine
			except Exception as e:
				print "ARE MATLAB AND MATLAB's PYTHON API INSTALLED CORRECTLY??"
				print e
				exit()
			for r in np.arange(options.rmin,options.rmax+0.0000000000001,options.rstep):
				#print("Radius:",r)
				idc+=1
				ressigma=lambdaengine.getResonancesWithSigmas(r,options.rtol)
				sigmaall+=ressigma
				self.addrowmatlab(options,ressigma,idc,r,options.rtol)
		print("DONE")

		'''
		if options.mfree>0:
			start_time=time.time()
			print("ADDING FREE RADIUSES")
			ressigma=np.arange(options.lambdamin,options.lambdamax+0.001,options.mfreestep)
			ressigma=[(round(l,1),0.0) for l in ressigma]
			#for r in np.arange(4,40+0.0000000000001,0.1):
			#	print("Radius:",r)
			#	ressigma+=lambdaengine.getResonancesWithSigmas(r,0.0)
			#ressigma=sorted(ressigma,key=itemgetter(0))
			#ressigma=[round(l,1) for (l,s) in ressigma]
			#ressigma=np.unique(ressigma)
			#ressigma=[(l,0.0) for l in ressigma]
			
			print "Free radius contains",len(ressigma),"lambda values"

			for m in range(0, options.mfree):
				print "Adding free radius #",m+1	
				idc+=1		
				self.addrowmatlab(options,ressigma,idc,-404-m,options.rtol)
			print("FREE RADIUS CREATION TIME: %.2f seconds" % (time.time() - start_time))
		'''
		#print self.aspinstance
		#exit()
		#lambdaengine.disposeEngine()



	def element(self, i, j):
		return self._rows['values'][i][j]

	def rows(self):
		return self._rows

	def row(self, r):
		return self._rows[r]

	def rowlen(self, r):
		return self._rows[r]['len']

	def nrows(self):
		return len(self._rows)

	def ncols(self):
		return self.sizebiggest

	def addrowmatlab(self,options,ressigma,idc,ray,tol, fromfile=False):
		newrow={}

		newrow['id']=int(idc)
		newrow['ray']=float(ray)
		newrow['tol']=float(tol)
		newrow['DL']=float(options.design_deltalambda)
		newrow['len']=len(ressigma)
		newrow['res']=ResonanceList()
		if ray>=0:
			self.aspinstance+="real"
		else:
			self.aspinstance+="special"
		self.aspinstance+="("+str(self.nrows()+1)+"). "
		

		index=0
		for (res,sigma) in ressigma:
			'''
			if options.design_uniformsigma>0:
				sigma=options.design_uniformsigma
			if options.design_disablesigma:
				sigma=0.0
			'''
			res=Resonance(res,sigma,str(self.nrows())+"_"+str(index),options.design_deltalambda, fromfile)
			self.aspinstance+="lambda("+str(self.nrows()+1)+","\
								+str(int(res.lb*100))+","\
								+str(int(res.val*100))+","\
								+str(int(res.ub*100))+"). "
			newrow['res'].append(res)
			index+=1
		self._rows.append(newrow)

		self.minr=min(newrow['res'].getRes()+[self.minr])
		self.maxr=max(newrow['res'].getRes()+[self.maxr])

		self.nelements+=len(newrow['res'].getRes())
		self.sizesmallest=min([len(newrow['res'].getRes()),self.sizesmallest])
		self.sizebiggest=max([len(newrow['res'].getRes()),self.sizebiggest])
		self.aspinstance+="\n"

	def addrow(self,row):
		rr=row.strip().split("\t")
		newrow={}
		newrow['id']=int(rr[1])
		newrow['ray']=float(rr[2])
		newrow['tol']=float(rr[3])
		newrow['DL']=float(rr[4])
		newrow['res']=ResonanceList()
		rr=rr[5].strip().split(":")
		newrow['len']=len(rr)/2
		if len(rr)%2!=0:
			raise FormatError()
		index=0
		for (res,sigma) in zip(*[iter(rr)]*2):
			if self.unisigma>0:
				sigma=self.unisigma
			if self.dissigma:
				sigma=0.0
			newrow['res'].append(Resonance(res,sigma,str(self.nrows())+"_"+str(index)))
			index+=1
		self._rows.append(newrow)

		self.minr=min(newrow['res'].getRes()+[self.minr])
		self.maxr=max(newrow['res'].getRes()+[self.maxr])

		self.nelements+=len(newrow['res'].getRes())
		self.sizesmallest=min([len(newrow['res'].getRes()),self.sizesmallest])
		self.sizebiggest=max([len(newrow['res'].getRes()),self.sizebiggest])

	def tofile(self, namefile):
		s=self.__str__()
		f=open(namefile,"w")
		f.write(s)
		f.close()

	def __str__(self, verbose=True):
		out="INSTANCE: NELEMENTS:"+str(self.nelements)+" NROWS:"+str(self.nrows())+"\n"
		out+="INSTANCE: UBlamda:"+str(self.UBlam)+"\n"
		out+="INSTANCE: LBlamda:"+str(self.LBlam)+"\n"
		#out+="INSTANCE: nu:"+str(self.nu)+"\n"
		#out+="INSTANCE: mu:"+str(self.mu)+" => K:"+str(self.K)+"\n"
		#out+="INSTANCE: k:"+str(len(self.nu))+"\n"
		#out+="INSTANCE: MINR:"+str(self.minr)+" MAXR:"+str(self.maxr)+" DIFF:"+str(self.maxr-self.minr)
		if not verbose:
			return out
		out+="\nID\tray\ttol\tDL\t#\tres1:sigma1\t...\n"
		for rs in self._rows:
			out+=str(rs['id'])+"\t"
			out+=str(rs['ray'])+"\t"
			out+=str(rs['tol'])+"\t"
			out+=str(rs['DL'])+"\t"
			out+=str(rs['len'])
			for d in rs['res'].reslist:
				out+="\t"+str(d.val)+":"+str(d.sigma)
			out+="\n"
		return out

	def get_rows_containing(self, notr, value, sigma=0):
			rr=[]
			for r in range(len(self._rows)):
				out=self._rows[r]['res'].get_interferences(value,sigma)
				if len(out)>0 and r!=notr:
					rr.append(r)
			return rr
			

	def get_values_within(self, value, sigma=0):
		out=[]
		for r in range(len(self._rows)):
			out+=self._rows[r]['res'].get_interferences(value,sigma)
		return out

	def get_row_from_node(self,node):
		r,e=node.split("_")
		r=int(r)
		return r

	def get_row_len(self,row):
		return self._rows[row]['res'].len()

	def allresonances(self):
		ress=[]
		for row in self._rows:
			ress+=row['res'].getRes()
		return list(set(ress))

	def values(r):
		return self._rows[r]['res']

	def arearay(r):
		return self._rows[r]['ray']*self._rows[r]['ray']*math.pi

	def get_value_from_node(self,node):	
		r,e=node.split("_")
		r=int(r)
		e=int(e)
		return self.rows()[r]['res'].getval(e)

	def get_sigma_from_node(self,node):	
		r,e=node.split("_")
		r=int(r)
		e=int(e)
		return self.rows()[r]['res'].getsigma(e)

	def get_nodes_with_value(self, value):
		nodes=[]
		for r in range(self.nrows()):
			ress=self.row(r)['res'].getRes()
			for n in range(len(ress)):
				if value==ress[n]:			
					nodes.append(str(r)+"_"+str(n))
					
		return nodes

	def get_ideal_span(self, k, p):
		if (p*k-1)==0:
			return 0
		return float(self.maxr-self.minr)/(p*k-1)

	def createOnlyNodes(self):
		print "CREATING GRAPH..."
		net=nx.DiGraph()
		for row in range(self.nrows()):
				for el in range(self.rows()[row]['len']):
					net.add_node(str(row)+"_"+str(el))
		return net

	def print_graph(self, insolution=[], spread=False):
		G=self.createOnlyNodes()
		insolution=[str(s[0])+"_"+str(s[1]) for s in insolution]
		
		A=nx.to_agraph(G)
		for n in A.nodes():
			nn=A.get_node(n)
			r,e=n.split("_")
			r=int(r)
			e=int(e)
			if spread:
				nn.attr['pos']="%f,%f"%(float(self.get_value_from_node(n)*100)-self.LBlam,float(len(self.rows())*200-r*200))
			else:
				nn.attr['pos']="%f,%f"%(float(e*200),float(len(self.rows())*200-r*200))
			nn.attr['fontsize']="50.0"
			nn.attr['shape']="none"

			if len(insolution)>0 and n in insolution:
				nn.attr['shape']="box"
				nn.attr['style']="rounded,filled,bold"
				nn.attr['fillcolor']="green"
			nn.attr['label']=self.get_value_from_node(n)

		sol=""
		if len(insolution)>0:
			sol="_sol"
		A.write(self.name+sol+".dot")
		A.draw(self.name+sol+".png",prog='neato',args='-n2')













