from gurobipy import *
from operator import itemgetter


class ResonanceModel(object):
	def __init__(self, net, instance, options):
		self.net=net
		self.inst=instance
		self.timelimit=options.milp_to
		self.threads=options.milp_threads
		self.k=options.milp_k
		self.p=options.milp_p
		self.d=options.milp_d
		self.cns_me=options.milp_uncns_me #interferences (mutual exclusion)
		self.cns_up=options.milp_uncns_up #uniform parallelism (intra radius cardinality)
		self.cns_rc=options.milp_uncns_rc #radius cardinality
		self.cns_pc=options.milp_uncns_pc #path cardinality
		#self.cns_disrsinc=options.disrsinc
		#self.cns_maxarea=options.maxarea
		self.minx=None
		self.x=[]		#X[e] tells whether the edge e is selected
		self.xp=[]		#xp[n] all X entering in the node n
		self.xm=[]		#xp[n] all X leaving from the node n
		self.eps={}		#EPS[n] tells whether n is selected or not
		self.epsrow={}	#epsrow[i] all EPS in the i-th radius
		self.y=[]		#Y[i] tells whether i-th radius is selected
		self.xd={}			#
		self.xplusrow={}	#all edges entering in the row
		self.xpluscol={}	#all edges entering in the column
		self.xminusrow={}	#all edges leaving from the row
		#self.v={}
		print
		print "----------------------"
		print "BUILDING MILP MODEL FOR MAX:DIST"
		self.__createvars()

		self.__impose_flow_constraints()

		if self.cns_rc:
			self.__impose_cardinality_constraints()
		if self.cns_up:
			self.__impose_uniformparallelism_constraints()
		if self.cns_me:
			self.__impose_mutualexcl_constraints()
			
		
		#if self.inst.N>0:
		#else:
		#	self.__impose_eteroparallelism_constraints()
	
		#if self.cns_maxarea>0:
		#	self.__impose_maxarea_constraints()
		#if not self.cns_disrsinc:
			#self.__impose_raysincompatibility_constraints()
		#	pass
		#self.model.update()
		#self.model.addConstr( self.model.getObjective() <= 18.12381,"boundOBJ")
		#self.model.update()
		print "DONE!"

	def __createvars(self):
		es=self.net.edges()
		ns=self.net.nodes()
		self.model = Model(self.inst.name)
		m=self.model
		m.params.threads=self.threads
		m.params.timelimit=self.timelimit

		for c in range(self.inst.ncols()):
			self.xpluscol[c]=[]

		for r in range(self.inst.nrows()):
			self.xplusrow[r]=[]
			self.xminusrow[r]=[]
			#self.v[r]=[]
			self.epsrow[r]=[]

		#self.minx=m.addVar(vtype=GRB.CONTINUOUS, obj=1, name="minx")
		#x vars
		for e in es:
			thisx=m.addVar(vtype=GRB.BINARY, obj=self.net.net.edge[e[0]][e[1]]['weight'], name="x"+str(e))
			self.x.append(thisx)
			self.xd[e]=thisx
			if e[0]!="S":
				self.xminusrow[int(self.net.get_row(e[0]))]+=[thisx]
			if e[1]!="T":
				self.xplusrow[int(self.net.get_row(e[1]))]+=[thisx]
				self.xpluscol[int(self.net.get_col(e[1]))]+=[thisx]
		
		m.update()

		#eps vars
		for n in ns:
			if n in ["T","S"]:
				continue
			self.eps[n]=m.addVar(vtype=GRB.BINARY, obj=0.0, name="eps"+str(n))
			self.epsrow[self.inst.get_row_from_node(n)].append(self.eps[n])

		#y vars
		for i in range(self.inst.nrows()):
			y=m.addVar(vtype=GRB.BINARY,name="y"+str(i))
			m.update()
			y.branchpriority=20
			self.y.append(y)

		#v vars
		'''
		for r in range(self.inst.nrows()):
			for h in range(self.inst.k):
				self.v[r].append(m.addVar(vtype=GRB.BINARY, obj=0, name="v"+str(r)+"_"+str(h)))
		'''
		m.update()

	def __str__(self):
		print "X SIZE:",len(self.x)
		print "Y SIZE:",len(self.y)
		print "EPS SIZE:",len(self.eps.items())
		print "V SIZE:",len(self.v.items())

	#it creates the dictionary of variables, indexed by edge
	def get_vars(self, bunche):
		return [self.xd[e] for e in bunche]

	#it returns the x variables of edges entering in n
	def get_star_plus(self, n):
		return self.get_vars(self.net.in_edges(n))

	#it returns the x variables of edges exiting from n
	def get_star_minus(self, n):
		return self.get_vars(self.net.out_edges(n))

	#it returns the x variables of edges entering in n's row
	def get_star_plus_row(self, n):
		return self.xplusrow[self.net.get_row(n)]

	#it returns the x variables of edges exiting from n's row
	def get_star_minus_row(self, n):
		return self.xminusrow[self.net.get_row(n)]

	def __impose_flow_constraints(self):
		print "ADDING FLOW CONSTRAINTS"
		#for x in self.x:
		#	self.model.addConstr( self.minx >= x*x.obj,"minxc")

		self.model.addConstr( quicksum(self.x) == self.k*self.p+1,"Length")

		if self.cns_pc:
			self.model.addConstr( quicksum(self.eps.values()) == self.k*self.p,"Length")

		for n in self.net.nodes():
			if n in ["T","S"]:
				continue
			self.model.addConstr( quicksum(self.get_star_plus(n)) == self.eps[n],"eps_"+n)

		for n in self.net.nodes():
			if n in ["T","S"]:
				continue
			self.model.addConstr( quicksum(self.get_star_plus(n))-quicksum(self.get_star_minus(n)) == 0,"Balance_"+n)
			#this is not mandatory, maybe improves solving
			#self.model.addConstr( quicksum(self.get_star_minus(n)) <= 1,"Maxout_"+n)
			
		self.model.addConstr( quicksum(self.get_star_plus("T")) == 1,"Balance_T")
		self.model.addConstr( quicksum(self.get_star_minus("S")) == 1,"Balance_S")

		#self.model.addConstr( quicksum(self.get_star_minus("S")) == 1,"Balance_S")
		#self.model.addConstr( self.eps["4_6"] ==1)
		'''
		for en in :
			self.model.addConstr( self.eps[en] ==1)
		'''
		#self.model.setObjective(self.minx)
		self.model.update()
		print "DONE"

	def __impose_cardinality_constraints(self):
		print "ADDING CARDINALITY CONSTRAINTS"
		#at most N elements per row can be selected
		for n in self.net.nodes():
			if n in ["T","S"]:
				continue
			self.model.addConstr( self.eps[n] <= self.y[self.inst.get_row_from_node(n)],"boundEpsY_"+str(n))
		self.model.addConstr( quicksum(self.y) == self.k,"boundY")
		self.model.update()
		print "DONE"

	def __impose_mutualexcl_constraints(self):
		print "ADDING ROWS' MUTUAL EXCLUSION CONSTRAINTS"
		everyrow=[]

		#for each selected element the rows containing that value must not be selected
		for n in self.net.nodes():
			if n in ["T","S"]:
				continue

			ri=self.inst.get_value_from_node(n)
			sigma=self.inst.get_sigma_from_node(n)
			row=self.inst.get_row_from_node(n)
			rows=self.inst.get_rows_containing(row, ri, sigma)
			#print n,rows
			rows=[rr for rr in rows if rr>row]
			if len(rows)==self.inst.nrows() and ri not in everyrow:
				everyrow.append(ri)
				print "WARNING: VALUE",ri,"IS IN EVERY ROW"
			exclude=[]

			for r in rows:
				exclude+=self.xplusrow[r]
				self.model.addConstr( self.eps[n] + self.y[r] <= 1,"exl_"+str(n)+"_r"+str(r))
		
		self.model.update()
		print "DONE"


	def __impose_uniformparallelism_constraints(self):
		print "ADDING UNIFORM PARALLELISM CONSTRAINTS"
		for r in range(self.inst.nrows()):
			self.model.addConstr( quicksum(self.epsrow[r]) == self.y[r]*self.p,"unpar_"+str(r) )
		self.model.update()
		print "DONE"
			
		
	def __impose_eteroparallelism_constraints(self):
		print "ADDING ETEROGENEOUS PARALLELISM CONSTRAINTS"
		for r in range(self.inst.nrows()):
			vvh=LinExp()
			for h in range(self.inst.k):
				self.model.addConstr( self.v[r][h] <= self.y[r],"etpar"+str(r))

			self.model.addConstr( quicksum(self.v[r]) == self.y[r],"etparY_"+str(r))
			self.model.addConstr( vvh.addTerms(self.nu[h],self.v[r][h]) == quicksum(self.epsrow[r]),"eparEPS_"+str(r))
		self.model.update()
		print "DONE"


	def __impose_maxarea_constraints(self):
		print "ADDING AREA CONSTRAINTS"
		vvh=LinExpr()
		for r in rows:
			for h in range(self.inst.k):
				vvh.add(self.v[r][h],self.mu[h]*self.inst.row(r)['ray'])

		self.model.addConstr( vvh <= self.cns_maxArea,"MAXAREA")
		print "DONE"			


	def from_varname_to_nodes(self,varname):
		n=""+varname
		n=n.replace("'","")
		n=n.replace("x(","")
		n=n.replace(")","")
		n=n.replace(", "," ")
		n=n.split(" ")
		return (n[0],n[1])


	def __impose_objective(self):
		self.model.modelSense = GRB.MINIMIZE

	def solve(self):
		self.model.optimize()
		if self.model.status == GRB.status.INFEASIBLE:
			print "MODEL IS INFEASIBLE"
			return "inf"
		if self.model.status == GRB.status.TIME_LIMIT:
			print "TIMELIMIT"
			if self.model.SolCount==0:
				return "to0"
			if self.model.SolCount==1:
				return "to1"
		if self.model.status == GRB.status.OPTIMAL:
			print "OPTIMAL"
			return "op"
			
		return True

	def get_solval(self):
		print "STORING SOLUTION"
		ress=[]
		yv=[yy.x for yy in self.y]
		ev=[]
		for e in sorted(self.eps.items(), key=itemgetter(0), reverse=True):
			ev.append(e[1].x)
		return (yv,ev)

	def set_startsol(self, yv, ev):
		print "SETTING INITIAL SOLUTION"
		self.model.update()
		ee=sorted(self.eps.items(), key=itemgetter(0), reverse=True)
		for i in range(0, len(ev)):
			ee[i][1].Start=ev[i]
		for i in range(0, len(yv)):
			self.y[i].Start=yv[i]
		self.model.update()
		
	def print_sol(self, allvar=True):
		ress=[]
		# Print solution
		print('\nTOTAL COST: %g' % self.model.objVal)
		print('IDEAL SPAN:'+str(self.inst.maxr)+"-"+str(self.inst.minr)+" / "+str((self.p*self.k)-1)+"="+str(self.net.ideal_span))
		print('SOLUTION:')
		for y in self.y:
			if round(y.x)>0:
				print y.getAttr("VarName"),y.x
		for e in sorted(self.eps.items(), key=itemgetter(0), reverse=True):
			if e[1].x>0:
				print "row",self.inst.get_row_from_node(e[0]),"res",self.inst.get_value_from_node(e[0])
				ress.append(self.inst.get_value_from_node(e[0]))
		ress=sorted(ress)
		print ress
		diffs=[ress[i]-ress[i-1] for i in range(1,len(ress))]
		print self.d,min(diffs),sum(diffs)*1.0/len(diffs),max(diffs)

	def get_edges_sol(self):
		edges=[]
		for r in range(self.inst.nrows()):
			for x in self.xplusrow[r]:
				if round(x.x) == 1.0:
					edges.append(self.from_varname_to_nodes(x.getAttr("VarName")))

			for x in self.xminusrow[r]:
				if round(x.x) == 1.0:
					edges.append(self.from_varname_to_nodes(x.getAttr("VarName")))
		return edges

	def get_nodes_sol(self):
		nodes=[]
		for n in self.eps.keys():
			if round(self.eps[n].x) == 1.0:
				nodes.append(n)
		return nodes
		
	def print_sol_graph(self):
		soledges=self.get_edges_sol()
		solenodes=self.get_nodes_sol()
		self.net.print_graph(self.inst, onlyedges=list(set(soledges)), onlynodes=solenodes)
























