from gurobipy import *
from operator import itemgetter


class ResonanceModel(object):
	def __init__(self, net, instance, options, maxpk=9999):
		self.net=net
		self.inst=instance
		self.timelimit=options.milp_to
		self.threads=options.milp_threads
		self.k=options.milp_k
		self.maxpk=maxpk
		self.p=options.milp_p
		self.d=options.milp_d
		self.cns_me=options.milp_uncns_me #interferences (mutual exclusion)
		self.cns_up=options.milp_uncns_up #uniform parallelism (intra radius cardinality)
		self.cns_rc=options.milp_uncns_rc #radius cardinality
		self.cns_pc=options.milp_uncns_pc #path cardinality
		#self.cns_disrsinc=options.disrsinc
		#self.cns_maxarea=options.maxarea
		self.MINC=None	#MINC[n] is the minimum parallelism
		self.EPS={}		#EPS[n] tells whether n is selected or not
		self.EPSr={}	#EPSr[i] all EPS in the i-th radius
		self.Y=[]		#Y[i] tells whether i-th radius is selected
		self.C=[]		#C[i] tells whether i-th radius is selected

		#self.v={}
		print
		print "----------------------"
		print "BUILDING MILP MODEL FOR MAX:P(k)"
		self.__createvars()

		if self.cns_rc:
			self.__impose_cardinality_constraints()
		if self.cns_me:
			self.__impose_mutualexcl_constraints()

		self.__impose_objective()
			
		
		#if self.inst.N>0:
		#else:
		#	self.__impose_eteroparallelism_constraints()
	
		#if self.cns_maxarea>0:
		#	self.__impose_maxarea_constraints()
		#if not self.cns_disrsinc:
			#self.__impose_raysincompatibility_constraints()
		#	pass
		#self.model.update()
		#self.model.addConstr( self.model.getObjective() <= 18.12381,"boundOBJ")
		#self.model.update()
		print "DONE!"

	def __createvars(self):
		es=self.net.edges()
		ns=self.net.nodes()
		self.model = Model(self.inst.name)
		m=self.model
		m.params.threads=self.threads
		m.params.timelimit=self.timelimit

		for r in range(self.inst.nrows()):
			self.EPSr[r]=[]

		#the last row of the table is the longest. However a better UB of the parallelism exists given k
		#that is the length of the -k-th radius from the end
		#self.MINC=m.addVar(vtype=GRB.CONTINUOUS, lb=0, ub=min(self.maxpk, self.inst.rowlen(-self.k)), obj=1, name="minc")
		self.MINC=m.addVar(vtype=GRB.CONTINUOUS, obj=1, name="minc")
		
		m.update()

		#eps vars
		for n in ns:
			if n in ["T","S"]:
				continue
			self.EPS[n]=m.addVar(vtype=GRB.BINARY, obj=0.0, name="eps"+str(n))
			self.EPSr[self.inst.get_row_from_node(n)].append(self.EPS[n])

		#y,c vars
		for i in range(self.inst.nrows()):
			y=m.addVar(vtype=GRB.BINARY,name="y"+str(i))
			m.update()
			y.branchpriority=20
			self.Y.append(y)

			#the upper bound of c(i) [maxpk comes from an iterative solving]
			c=m.addVar(vtype=GRB.INTEGER, lb=0, ub=min(self.maxpk,self.inst.rowlen(i)), name="C"+str(i))
			self.C.append(c)

		m.update()

	def __str__(self):
		print "Y SIZE:",len(self.Y)
		print "C SIZE:",len(self.C)
		print "EPS SIZE:",len(self.EPS.items())

	def __impose_cardinality_constraints(self):
		print "ADDING CARDINALITY CONSTRAINTS"
		#at most N elements per row can be selected
		for n in self.net.nodes():
			if n in ["T","S"]:
				continue
			self.model.addConstr( self.EPS[n] <= self.Y[self.inst.get_row_from_node(n)],"boundEpsY_"+str(n))

		self.model.addConstr( quicksum(self.Y) == self.k,"boundY")

		for i in range(self.inst.nrows()):
			self.model.addConstr( quicksum(self.EPSr[i]) == self.C[i],"boundC0")
			self.model.addConstr( self.Y[i] <= self.C[i],"boundC1")
			self.model.addConstr( self.Y[i]*self.maxpk >= self.C[i],"boundC2")
			self.model.addConstr( self.MINC <= (1-self.Y[i])*1000+self.C[i],"objc")

		self.model.update()
		self.model.write("djdjdd.lp")
		print "DONE"

	def __impose_mutualexcl_constraints(self):
		print "ADDING ROWS' MUTUAL EXCLUSION CONSTRAINTS"
		everyrow=[]

		#for each selected element the rows containing that value must not be selected
		for n in self.net.nodes():
			if n in ["T","S"]:
				continue

			ri=self.inst.get_value_from_node(n)
			sigma=self.inst.get_sigma_from_node(n)
			row=self.inst.get_row_from_node(n)
			rows=self.inst.get_rows_containing(row, ri, sigma)
			print n,ri,rows
			rows=[rr for rr in rows]# if rr>row]
			if len(rows)==self.inst.nrows() and ri not in everyrow:
				everyrow.append(ri)
				print "WARNING: VALUE",ri,"IS IN EVERY ROW"

			for r in rows:
				self.model.addConstr( self.EPS[n] + self.Y[r] <= 1,"exl_"+str(n)+"_r"+str(r))
		
		self.model.update()
		print "DONE"

	def __impose_objective(self):
		self.model.modelSense = GRB.MAXIMIZE

	def solve(self):
		self.model.optimize()
		if self.model.status == GRB.status.INFEASIBLE:
			print "MODEL IS INFEASIBLE"
			return "inf"
		if self.model.status == GRB.status.TIME_LIMIT:
			print "TIMELIMIT"
			if self.model.SolCount==0:
				return "to0"
			if self.model.SolCount==1:
				return "to1"
		if self.model.status == GRB.status.OPTIMAL:
			print "OPTIMAL"
			return "op"
			
		return True

	def get_solval(self):
		print "STORING SOLUTION"
		ress=[]
		yv=[yy.x for yy in self.Y]
		ev=[]
		for e in sorted(self.EPS.items(), key=itemgetter(0), reverse=True):
			ev.append(e[1].x)
		return (yv,ev)

	def get_solbottompar(self):
		bottom=99999
		for i in range(self.inst.nrows()):
			if round(self.Y[i].x)==1:
				bottom=min(bottom, sum([e.x for e in self.EPSr[i]]))
		return bottom

	def set_startsol(self, yv, ev):
		print "SETTING INITIAL SOLUTION"
		self.model.update()
		ee=sorted(self.eps.items(), key=itemgetter(0), reverse=True)
		for i in range(0, len(ev)):
			ee[i][1].Start=ev[i]
		for i in range(0, len(yv)):
			self.y[i].Start=yv[i]
		self.model.update()
		
	def print_sol(self, allvar=True):
		ress=[]
		resdict={}
		# Print solution
		print('\nTOTAL COST: %g' % self.model.objVal)
		print('IDEAL SPAN:'+str(self.inst.maxr)+"-"+str(self.inst.minr)+" / "+str((self.p*self.k)-1)+"="+str(self.net.ideal_span))
		print('SOLUTION:')
		for y in self.Y:
			if round(y.x)>0:
				resdict[int(y.getAttr("VarName")[1:])]=[]
				print y.getAttr("VarName"),y.x
		print resdict
		for e in sorted(self.EPS.items(), key=itemgetter(0), reverse=True):
			if e[1].x>0:
				resdict[self.inst.get_row_from_node(e[0])].append(self.inst.get_value_from_node(e[0]))
				#print "row",self.inst.get_row_from_node(e[0]),"res",self.inst.get_value_from_node(e[0])
				ress.append(self.inst.get_value_from_node(e[0]))
		for kk in resdict.keys():
			print "row ", kk, " selected:", len(resdict[kk]), " list:", sorted(resdict[kk]) 
		ress=sorted(ress)
		print ress
		if len(ress)==1:
			diffs=[0]
		else:
			diffs=[ress[i]-ress[i-1] for i in range(1,len(ress))]
		print min(diffs),sum(diffs)*1.0/len(diffs),max(diffs)

	def get_nodes_sol(self):
		nodes=[]
		for n in self.EPS.keys():
			if round(self.EPS[n].x) == 1.0:
				nodes.append(n)
		return nodes
























